""" QCheckboxConfig module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Callable, Dict
from functools import partial
from PyQt5.QtWidgets import QAction, QMenu
from magneticalc.Debug import Debug


class QCheckboxConfig:
    """ QCheckboxConfig class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Initializes a manager for configuration bound checkbox actions.
        """
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.actions: Dict[str, Dict] = {}

    def add(self, label: str, key: str, on_changed: Callable) -> QAction:
        """
        Creates a checkbox action.
        The checkbox state is bound to the project configuration.

        @param label: Label
        @param key: Configuration key
        @param on_changed: Gets called when the checkbox state changed
        """
        action = QAction(label)
        action.setCheckable(True)
        action.triggered.connect(  # type: ignore
            partial(self.changed, key)
        )
        self.actions[key] = {"action": action, "on_changed": on_changed}
        action.setChecked(self.gui.project.get_bool(key))
        return action

    def changed(self, key: str) -> None:
        """
        Handles change of checkbox state.

        @param key: Configuration key
        """
        self.gui.project.set_bool(key, self.actions[key]["action"].isChecked())
        self.actions[key]["on_changed"]()

    def reload(self) -> None:
        """
        Reloads the configuration bound checkboxes.

        Note: This doesn't block the change signals.
        """
        for item in self.actions.items():
            key, dictionary = item
            dictionary["action"].setChecked(self.gui.project.get_bool(key))



""" Wire module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Dict, Callable
import numpy as np
from PyQt5.QtCore import QThread
from magneticalc.Assert_Dialog import Assert_Dialog
from magneticalc.Debug import Debug
from magneticalc.Validatable import Validatable, require_valid, validator


class Wire(Validatable):
    """ Wire class. """

    def __init__(self) -> None:
        """
        Initializes an empty wire.
        A 3D piecewise linear curve with some DC current associated with it.
        """
        Validatable.__init__(self)
        Debug(self, ": Init", init=True)

        self.points_base = np.array([])
        self.points_transformed = np.array([])

        """ Bounding box: Minimum bounds (3D point), maximum bounds (3D point). """
        self.bounds = np.array([]), np.array([])

        self._slicer_limit = 0
        self.dc = 0

        self._points_sliced = np.array([])
        self._length = 0

        """ Wire Elements: Ordered list of segment center points and directions. """
        self._elements = np.array([])

    def set(
            self,
            points: np.ndarray,
            stretch: np.ndarray,
            rotational_symmetry: Dict,
            close_loop: bool,
            slicer_limit: float,
            dc: float
    ) -> None:
        """
        Sets the parameters.

        @param points: Ordered list of 3D coordinates (see presets)
        @param stretch: XYZ stretch transform factors (3D point)
        @param rotational_symmetry: Dictionary for rotational symmetry transform
        @param close_loop: Enable to transform the wire into a closed loop (append first point)
        @param slicer_limit: Slicer limit
        @param dc: Wire current (A)
        """
        self._slicer_limit = slicer_limit
        self.dc = dc

        self._points_sliced = np.array([])
        self._length = 0

        self.points_base = np.array(points)

        axes = np.array(points).T
        axes = self._transform_stretch(axes, stretch)
        axes = self._transform_rotational_symmetry(axes, rotational_symmetry, close_loop=close_loop)
        self.points_transformed = axes.T

        self.bounds = [min(axes[0]), min(axes[1]), min(axes[2])], [max(axes[0]), max(axes[1]), max(axes[2])]

        Assert_Dialog(len(self.points_base) >= 2, "Number of points must be >= 2")

    @staticmethod
    def _transform_stretch(axes: np.ndarray, stretch: np.ndarray) -> np.ndarray:
        """
        This transformation stretches (and/or mirrors) this curve by some factor in any direction.
        Use the factor +1 / -1 to retain / mirror the curve in that direction.

        @param axes: Transposed array of Wire points
        @param stretch: XYZ stretch transform factors (3D point)
        @return: Transposed array of Wire points
        """
        for i in range(3):
            axes[i] *= stretch[i]
        return axes

    @staticmethod
    def _transform_rotational_symmetry(axes: np.ndarray, parameters: Dict, close_loop: bool) -> np.ndarray:
        """
        This transformation replicates and rotates this curve `count` times about an `axis` with radius `radius`.

        @param axes: Transposed array of Wire points
        @param parameters: Dictionary containing the transformation parameters
                           (number of replications, radius, axis and offset angle)
        @param close_loop: Enable to transform the wire into a closed loop (append first point)
        @return: Transposed array of Wire points
        """
        x, y, z = [], [], []

        axis_other_1 = (parameters["axis"] + 1) % 3
        axis_other_2 = (parameters["axis"] + 2) % 3

        for a in np.linspace(0, 2 * np.pi, parameters["count"], endpoint=False):
            b = a + parameters["offset"] * np.pi / 180
            x = np.append(x, axes[axis_other_1] * np.sin(b) - (axes[axis_other_2] + parameters["radius"]) * np.cos(b))
            y = np.append(y, axes[axis_other_1] * np.cos(b) + (axes[axis_other_2] + parameters["radius"]) * np.sin(b))
            z = np.append(z, axes[parameters["axis"]])

        axes = [x, y, z]

        if close_loop:
            for i in range(3):
                axes[i] = np.append(axes[i], axes[i][0])

        return np.array(axes)

    # ------------------------------------------------------------------------------------------------------------------

    @validator
    def recalculate(self, progress_callback: Callable) -> bool:
        """
        Slices wire segments into smaller ones until segment lengths equal or undershoot slicer limit.

        @param progress_callback: Progress callback
        @return: True if successful, False if interrupted
        """
        Debug(self, ".recalculate()")

        points_sliced = []
        length = 0

        for i in range(len(self.points_transformed) - 1):

            # Calculate direction and length of wire segment
            segment_direction = np.array(self.points_transformed[i + 1] - self.points_transformed[i])
            segment_length = np.linalg.norm(segment_direction)
            length += segment_length

            # Calculate required number of slices (subdivisions) and perform linear interpolation
            slices = np.ceil(segment_length / self._slicer_limit).astype(int)
            linear = np.linspace(0, 1, slices, endpoint=False)
            points_sliced += [self.points_transformed[i] + segment_direction * j for j in linear]

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                progress_callback(50 * (i + 1) / (len(self.points_transformed) - 1))

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".recalculate(): WARNING: Interruption requested, exiting now", warning=True)
                    return False

        # Append the very last point since it is not appended by the interpolation above
        points_sliced.append(self.points_transformed[-1])

        self._points_sliced = np.array(points_sliced)
        self._length = length

        # Calculate wire elements: [[element_center, element_direction], …]
        elements = []
        for i in range(len(self._points_sliced) - 1):

            element_direction = np.array(self._points_sliced[i + 1]) - np.array(self._points_sliced[i])
            element_center = self._points_sliced[i] + element_direction / 2
            elements.append([element_center, element_direction])

            # Signal progress update, handle interrupt (every 16 iterations to keep overhead low)
            if i & 0xf == 0:
                progress_callback(50 + 50 * (i + 1) / (len(self._points_sliced) - 1))

                if QThread.currentThread().isInterruptionRequested():
                    Debug(self, ".recalculate(): WARNING: Interruption requested, exiting now", warning=True)
                    return False

        self._elements = np.array(elements)

        progress_callback(100)

        return True

    @property
    @require_valid
    def points_sliced(self) -> np.ndarray:
        """
        Returns this wire's points after slicing.

        @return: Ordered list of 3D points
        """
        return self._points_sliced

    @property
    @require_valid
    def length(self) -> float:
        """
        @return: Wire length
        """
        return self._length

    @property
    @require_valid
    def elements(self) -> np.ndarray:
        """
        @return: Wire elements
        """
        return self._elements



""" Model module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional, Dict, Callable
from sty import fg
from magneticalc.Debug import Debug
from magneticalc.Field import Field
from magneticalc.Field_Types import FIELD_TYPE_A, FIELD_TYPE_B
from magneticalc.Metric import Metric
from magneticalc.ModelAccess import ModelAccess
from magneticalc.Parameters import Parameters
from magneticalc.SamplingVolume import SamplingVolume
from magneticalc.Wire import Wire


class Model:
    """
    Model class.

    The model maintains a strict hierarchy of dependencies:  parameters > metric > field > sampling volume > wire
    When a lower module's data changed, all higher modules are invalidated (i.e. have their calculation results reset).
    """

    # Used by L{Debug}
    DebugColor = fg.yellow

    """ Enable to debug calls to L{invalidate()}. """
    DebugInvalidate = False

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Initializes the model.

        @param gui: GUI
        """
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.wire = Wire()
        self.sampling_volume = SamplingVolume()
        self.metric = Metric()
        self.parameters = Parameters()

        self._field_type_select: int = FIELD_TYPE_A
        self._field_cache: Dict[int, Field] = {
            FIELD_TYPE_A: Field(),
            FIELD_TYPE_B: Field(),
        }

    @property
    def field_type_select(self) -> int:
        """
        @return: Currently selected field type
        """
        return self._field_type_select

    @field_type_select.setter
    def field_type_select(self, field_type: int) -> None:
        """
        Sets the currently selected field type.

        @param field_type: Field type
        """
        Debug(self, f".field_type_select = {field_type}")
        self._field_type_select = field_type

    @property
    def field(self) -> Field:
        """
        @return: Currently selected field
        """
        return self._field_cache[self.field_type_select]

    def get_valid_field(self, field_type: int) -> Optional[Field]:
        """
        Gets a field by type if the field is valid.

        @param field_type: Field type
        @return: Field if valid, None otherwise
        """
        field = self._field_cache[field_type]
        return field if field.valid else None

    # ------------------------------------------------------------------------------------------------------------------

    @property
    def valid(self) -> bool:
        """
        @return: True if model is valid, False otherwise
        """
        return \
            self.wire.valid and \
            self.sampling_volume.valid and \
            self.field.valid and \
            self.metric.valid and \
            self.parameters.valid

    def invalidate(
            self,
            do_wire: bool = False,
            do_sampling_volume: bool = False,
            do_field: bool = False,
            do_metric: bool = False,
            do_parameters: bool = False,
            do_all: bool = False
    ) -> None:
        """
        Invalidates multiple hierarchy levels at once, in descending order of hierarchy.

        @param do_wire: Enable to invalidate wire
        @param do_sampling_volume: Enable to invalidate sampling volume
        @param do_field: Enable to invalidate all fields
        @param do_metric: Enable to invalidate metric
        @param do_parameters: Enable to invalidate parameters
        @param do_all: Enable to invalidate all hierarchy levels
        """
        if self.DebugInvalidate:
            subject = {
                "wire": do_wire,
                "sampling_volume": do_sampling_volume,
                "field": do_field,
                "metric": do_metric,
                "parameters": do_parameters,
                "all": do_all
            }
            string = ", ".join([name for name, condition in subject.items() if condition])
            Debug(self, f".invalidate({string})")

        if do_all or do_parameters:
            if self.parameters.valid:
                self.parameters.valid = False
                self.on_parameters_invalid()

        if do_all or do_metric:
            if self.metric.valid:
                self.metric.valid = False
                self.on_metric_invalid()

        if do_all or do_field:
            valid_fields = [field for field_type, field in self._field_cache.items() if field.valid]
            if any(valid_fields):
                for field in valid_fields:
                    field.valid = False
                self.on_field_invalid()

        if do_all or do_sampling_volume:
            if self.sampling_volume.valid:
                self.sampling_volume.valid = False
                self.on_sampling_volume_invalid()

        if do_all or do_wire:
            if self.wire.valid:
                self.wire.valid = False
                self.on_wire_invalid()

    def set_wire(
            self,
            invalidate: bool,
            *args,
            **kwargs
    ) -> None:
        """
        Sets the wire.

        @param invalidate: Enable to invalidate this model hierarchy level
        """
        Debug(self, ".set_wire()")

        with ModelAccess(self.gui, recalculate=False):

            self.invalidate(
                do_wire=invalidate,
                do_sampling_volume=True,
                do_field=True,
                do_metric=True,
                do_parameters=True
            )
            self.wire.set(*args, **kwargs)

    def set_sampling_volume(
            self,
            invalidate: bool,
            *args,
            **kwargs
    ) -> None:
        """
        Sets the sampling volume.

        @param invalidate: Enable to invalidate this model hierarchy level
        """
        Debug(self, ".set_sampling_volume()")

        with ModelAccess(self.gui, recalculate=False):

            self.invalidate(
                do_sampling_volume=invalidate,
                do_field=True,
                do_metric=True,
                do_parameters=True
            )
            self.sampling_volume.set(*args, **kwargs)

    def set_field(
            self,
            invalidate: bool,
            field_type: int,
            *args,
            **kwargs
    ) -> None:
        """
        Sets the field.

        @param invalidate: Enable to invalidate this model hierarchy level
        @param field_type: Field type
        """
        Debug(self, ".set_field()")

        with ModelAccess(self.gui, recalculate=False):

            self.field_type_select = field_type

            self.invalidate(
                do_field=invalidate,
                do_metric=True,
                do_parameters=True
            )

            self.field.set(type=field_type, *args, **kwargs)

    def set_metric(
            self,
            invalidate: bool,
            *args,
            **kwargs
    ) -> None:
        """
        Sets the metric.

        @param invalidate: Enable to invalidate this model hierarchy level
        """
        Debug(self, ".set_metric()")

        with ModelAccess(self.gui, recalculate=False):

            self.invalidate(
                do_metric=invalidate,
                do_parameters=True
            )
            self.metric.set(*args, **kwargs)

    def set_parameters(
            self,
            invalidate: bool,
            *args,
            **kwargs
    ) -> None:
        """
        Sets the parameters.

        @param invalidate: Enable to invalidate this model hierarchy level
        """
        Debug(self, ".set_parameters()")

        with ModelAccess(self.gui, recalculate=False):

            self.invalidate(
                do_parameters=invalidate
            )
            self.parameters.set(*args, **kwargs)

    # ------------------------------------------------------------------------------------------------------------------

    def calculate_wire(self, progress_callback: Callable) -> None:
        """
        Calculates the wire.

        @param progress_callback: Progress callback
        @return: True if successful, False if interrupted
        """
        Debug(self, ".calculate_wire()")
        self.invalidate(do_sampling_volume=True, do_field=True, do_metric=True)
        return self.wire.recalculate(progress_callback=progress_callback)

    def calculate_sampling_volume(self, progress_callback: Callable) -> bool:
        """
        Calculates the sampling volume.

        @param progress_callback: Progress callback
        @return: True if successful, False if interrupted
        """
        Debug(self, ".calculate_sampling_volume()")
        self.invalidate(do_field=True, do_metric=True)
        return self.sampling_volume.recalculate(progress_callback=progress_callback)

    def calculate_field(self, progress_callback: Callable, num_cores: int, backend_type: int) -> bool:
        """
        Calculates the field.

        @param progress_callback: Progress callback
        @param num_cores: Number of CPU cores to use
        @param backend_type: Backend type
        @return: True if successful, False if interrupted
        """
        Debug(self, f".calculate_field(num_cores={num_cores})")
        self.invalidate(do_metric=True)
        return self.field.recalculate(
            wire=self.wire,
            sampling_volume=self.sampling_volume,
            progress_callback=progress_callback,
            num_cores=num_cores,
            backend_type=backend_type
        )

    def calculate_metric(self, progress_callback: Callable) -> bool:
        """
        Calculates the metric.

        @param progress_callback: Progress callback
        @return: True (currently non-interruptable)
        """
        Debug(self, ".calculate_metric()")
        return self.metric.recalculate(
            sampling_volume=self.sampling_volume,
            field=self.field,
            progress_callback=progress_callback
        )

    def calculate_parameters(self, progress_callback: Callable) -> bool:
        """
        Calculates the parameters.

        @param progress_callback: Progress callback
        @return: True (currently non-interruptable)
        """
        Debug(self, ".calculate_parameters()")
        return self.parameters.recalculate(
            wire=self.wire,
            sampling_volume=self.sampling_volume,
            field=self.field,
            progress_callback=progress_callback
        )

    # ------------------------------------------------------------------------------------------------------------------

    def on_wire_valid(self) -> None:
        """
        Gets called when the wire was successfully calculated.
        """
        Debug(self, ".on_wire_valid()")

        self.gui.sidebar_left.wire_widget.update()
        self.gui.menu.wire_menu.update()
        self.gui.vispy_canvas.redraw()

    def on_sampling_volume_valid(self) -> None:
        """
        Gets called when the sampling volume was successfully calculated.
        """
        Debug(self, ".on_sampling_volume_valid()")

        self.gui.sidebar_left.sampling_volume_widget.update()
        self.gui.sidebar_right.display_widget.update()
        self.gui.sidebar_right.display_widget.prevent_excessive_field_labels(choice=False)
        self.gui.vispy_canvas.redraw()

    def on_field_valid(self) -> None:
        """
        Gets called when the field was successfully calculated.
        """
        Debug(self, ".on_field_valid()")

        self.gui.sidebar_right.field_widget.update()
        self.gui.vispy_canvas.redraw()

    def on_metric_valid(self) -> None:
        """
        Gets called when the metric was successfully calculated.

        Note: Field label creation is triggered on-demand inside L{VisPyCanvas.redraw()}, not here.
        """
        Debug(self, ".on_metric_valid()")

        self.gui.sidebar_right.metric_widget.update()
        self.gui.vispy_canvas.redraw()

    def on_parameters_valid(self) -> None:
        """
        Gets called when the parameters were successfully calculated.
        """
        Debug(self, ".on_parameters_valid()")

        self.gui.sidebar_right.parameters_widget.update()
        self.gui.vispy_canvas.redraw()

    # ------------------------------------------------------------------------------------------------------------------

    def on_wire_invalid(self) -> None:
        """
        Gets called when the wire was invalidated.
        """
        Debug(self, ".on_wire_invalid()")

        self.gui.sidebar_left.wire_widget.update()
        self.gui.menu.wire_menu.update()

    def on_sampling_volume_invalid(self) -> None:
        """
        Gets called when the sampling volume was invalidated.
        """
        Debug(self, ".on_sampling_volume_invalid()")

        self.gui.sidebar_left.sampling_volume_widget.update()

    def on_field_invalid(self) -> None:
        """
        Gets called when the field was invalidated.
        """
        Debug(self, ".on_field_invalid()")

        self.gui.sidebar_right.field_widget.update()

    def on_metric_invalid(self) -> None:
        """
        Gets called when the metric was invalidated.
        """
        Debug(self, ".on_metric_invalid()")

        self.gui.sidebar_right.metric_widget.update()
        self.gui.vispy_canvas.delete_field_labels()

    def on_parameters_invalid(self) -> None:
        """
        Gets called when the parameters were invalidated.
        """
        Debug(self, ".on_parameters_invalid()")

        self.gui.sidebar_right.parameters_widget.update()



""" QLayouted module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Dict, Tuple, Callable, Union, Optional
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QLayout, QWidget, QPushButton
from magneticalc.QtWidgets2.QButtons import QButtons


class QLayouted:
    """ QLayouted class. """

    def __init__(self, direction: str = "vertical") -> None:
        """
        Initializes the QLayouted class.
        This adds a layout and several related functions like addWidget() to the parent class.

        @param direction: Sets "vertical" or "horizontal" layout
        """
        self._layout = QVBoxLayout() if direction == "vertical" else QHBoxLayout()

    def install_layout(self, parent: QWidget) -> None:
        """
        Installs this layout in the parent.
        """
        parent.setLayout(self._layout)

    # noinspection PyPep8Naming
    def addWidget(self, widget, alignment: Optional[Union[Qt.Alignment, Qt.AlignmentFlag]] = None) -> None:
        """
        Adds widget.

        @param widget: QWidget
        @param alignment: Alignment
        """
        if alignment:
            self._layout.addWidget(widget)
        else:
            self._layout.addWidget(widget)

    # noinspection PyPep8Naming
    def addLayout(self, layout: QLayout) -> None:
        """
        Adds layout.

        @param layout: QLayout
        """
        self._layout.addLayout(layout)

    # noinspection PyPep8Naming
    def addSpacing(self, spacing: int) -> None:
        """
        Adds spacing.

        @param spacing: Spacing value
        """
        self._layout.addSpacing(spacing)

    # noinspection PyPep8Naming
    def addButtons(self, data: Dict[str, Tuple[str, Callable]]) -> Dict[int, QPushButton]:
        """
        Adds buttons.

        @param data: Dictionary {text: (icon, callback), …}
        @return: Dictionary {index: QPushButton, …}
        """
        buttons = QButtons(data)
        self.addLayout(buttons)
        return buttons.dictionary



""" MagnetiCalc setup script. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from setuptools import setup, find_packages

version = {}
with open("magneticalc/Version.py") as file:
    exec(file.read(), version)

setup(
    packages=find_packages(),
    name="MagnetiCalc",
    version=version["__VERSION__"],
    description="""MagnetiCalc calculates the magnetic flux density, vector potential, energy, self-inductance
and magnetic dipole moment of arbitrary coils. Inside an OpenGL-accelerated GUI,
the static magnetic flux density (B-field) or the magnetic vector potential (A-field)
is displayed in interactive 3D, using multiple metrics for highlighting the field properties.""".replace("\n", " "),
    url=version["__URL__"],
    author="Paul Wilhelm, M. Sc.",
    author_email="anfrage@paulwilhelm.de",
    license="ISC License",
    classifiers=[
        "License :: OSI Approved :: ISC License (ISCL)",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Development Status :: 4 - Beta"
    ],
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    platforms=["any"],
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "numba>=0.50.0",
        "scipy",
        "PyQt5",
        "vispy>=0.6.5",
        "qtawesome>=1.0.0,<=1.3.1",
        "sty",
        "si-prefix>=1.2.2",
        "h5py"
    ]
)



""" QTableView2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Any, Optional, List, Dict, Callable
from functools import partial
from PyQt5 import QtCore
from PyQt5.Qt import QFocusEvent
from PyQt5.QtCore import Qt, QItemSelectionModel, QItemSelection, QAbstractTableModel, QVariant
from PyQt5.QtWidgets import QTableView, QHeaderView, QAbstractItemView, QComboBox, QSizePolicy
from magneticalc.QtWidgets2.QPushButton2 import QPushButton2
from magneticalc.Debug import Debug
from magneticalc.Format import Format
from magneticalc.Theme import Theme


class TableModel(QAbstractTableModel):
    """ TableModel class. """

    def __init__(
            self,
            parent: QTableView,
            data: List[List],
            row_keys: List[str],
            col_keys: List[str],
            col_is_widget: Optional[List[bool]] = None
    ) -> None:
        """
        Initializes a table model.

        @param parent: QTableView
        @param data: Data (2D)
        @param row_keys: List of vertical header strings
        @param col_keys: List of horizontal header strings
        @param col_is_widget: Indicates columns contains only widgets instead of only data (may be None for data-only)
        """
        QAbstractTableModel.__init__(self, parent=parent)
        self._data = data
        self._row_keys = row_keys
        self._col_keys = col_keys
        self._col_is_widget = col_is_widget

    def is_widget(self, index: QtCore.QModelIndex) -> bool:
        """
        Checks if the cell is a widget button.

        @param index: Cell index
        @return: True if the cell is a widget, False otherwise
        """
        return False if self._col_is_widget is None else self._col_is_widget[index.column()]

    def rowCount(self, parent: QtCore.QModelIndex = None):
        """
        Gets the number of rows.

        @return: Number of rows
        """
        return len(self._data)

    def columnCount(self, parent: QtCore.QModelIndex = None):
        """
        Gets the number of columns.
        Note: Enabling allow_delete adds an extra column for the row delete buttons.

        @return: Number of columns
        """
        return len(self._data[0]) if self.rowCount() > 0 else 0

    def setData(self, index: QtCore.QModelIndex, value: Any, role: int = Qt.EditRole) -> bool:
        """
        Sets a cell value.

        @param index: Cell index
        @param value: Cell value
        @param role: Data role
        @return: True if successful, False otherwise
        """
        if role == Qt.EditRole:
            if not self.is_widget(index):
                self._data[index.row()][index.column()] = value
                self.dataChanged.emit(index, index)  # type: ignore
                return True
        return False

    def data(self, index: QtCore.QModelIndex, role: int = Qt.DisplayRole) -> QVariant:
        """
        Gets a cell value.

        @param index: Cell index
        @param role: Data role
        @return: Cell value
        """
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if not self.is_widget(index):
                return self._data[index.row()][index.column()]
        return QVariant()

    def flags(self, index: QtCore.QModelIndex) -> int:
        """
        Gets the flags for a cell.

        @param index: Cell index
        @return: Cell flags
        """
        if self.is_widget(index):
            return Qt.ItemIsEnabled
        else:
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsEditable

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> QVariant:
        """
        Sets the header data.

        @param section: Row or column
        @param orientation: Orientation
        @param role: Data role
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return QVariant(self._col_keys[section])
        elif orientation == Qt.Vertical and role == Qt.DisplayRole:
            return QVariant(self._row_keys[section])
        return QVariant()


class QTableView2(QTableView):
    """ QTableView2 class. """

    def __init__(
            self,
            gui: GUI,  # type: ignore
            col_keys: List[str],
            col_types: Optional[Dict] = None,
            col_options: Optional[List] = None,
            row_allow_delete: bool = True,
            row_count_minimum: int = 0,
            row_prefix: str = "",
            on_cell_edited: Optional[Callable] = None,
            on_selection_changed: Optional[Callable] = None,
            on_row_deleted: Optional[Callable] = None
    ) -> None:
        """
        Initializes a table.

        @param gui: GUI
        @param col_keys: List of horizontal header strings
        @param col_types: Per-column cell types used for binding cells to configuration.
                          Ordered dictionary of key-value pairs (column key : column type) or None to disable binding.
                          Note: Currently, *only the keys* are used for binding cells to project configuration;
                          the cell options actually determine the cell type, i.e. numerical or combobox.
        @param col_options: Per-column cell options. This determines the cell type, i.e. numerical or combobox.
                            List of options for every column; items may be set to None to use numerical cells.
        @param row_allow_delete: Enable to add an extra column for row delete buttons
        @param row_count_minimum: Minimum number of rows (no further rows can be deleted)
        @param row_prefix: Per-row prefix used for binding cells to configuration
                           (key = prefix + column key + "_" + row index)
        @param on_cell_edited: Set this to make cells editable
        @param on_selection_changed: Used to inform the GUI that another row was selected
        @param on_row_deleted: Set this to make rows deletable
        """
        QTableView.__init__(self)
        Debug(self, ": Init", init=True)
        self.gui = gui

        self._allow_delete = row_allow_delete
        self._row_count_minimum = row_count_minimum
        self._row_prefix = row_prefix
        self._col_keys = col_keys
        self._col_types = col_types
        self._col_options = col_options
        self._on_cell_edited = on_cell_edited
        self._on_selection_changed = on_selection_changed
        self._on_row_deleted = on_row_deleted

        self.setFocusPolicy(Qt.StrongFocus)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.set_style(border_color="black", border_width=1)
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectItems)

    def set_style(self, border_color: str, border_width: int) -> None:
        """
        Sets this widget's style.

        @param border_color: Border color
        @param border_width: Border width
        """
        self.setStyleSheet(f"""
            QTableView {{
                border: {border_width}px solid {border_color};
                border-radius: 5px;
                padding: 1px;
                margin: {2 - border_width}px;
            }}

            QTableCornerButton::section {{
                border-width: 1px;
                border-style: solid;
                border-color: white silver silver white;
            }}

            QTableView::item {{
                padding: 2;
            }}
        """)

    # ------------------------------------------------------------------------------------------------------------------

    def set_data(
            self,
            data: List[List[str]],
            row_keys: List[str]
    ) -> None:
        """
        Sets the table contents.

        @param data: Data (2D)
        @param row_keys: List of vertical header strings
        """
        Debug(self, ".set_data()")

        if self.model() is not None:
            # Disconnect previously set up callbacks
            if self._on_cell_edited is not None:
                self.model().dataChanged.disconnect(self.on_data_changed)  # type: ignore
            if self._on_selection_changed is not None:
                self.selectionModel().selectionChanged.disconnect(self.on_selection_changed)  # type: ignore

        # Initialize cells that are bound to the project configuration
        if self._col_types is not None:
            for row, row_data in enumerate(data):
                for col, col_data in enumerate(data):
                    key = list(self._col_types.keys())[col]
                    data[row][col] = self.gui.project.get_str(self._row_prefix + key + "_" + str(row))

        # Indicate if a column contains only widgets (only data otherwise)
        if self._col_options is not None:
            col_is_widget = [col_options is not None for col_options in self._col_options] + \
                ([True] if self._allow_delete else [])
        else:
            col_is_widget = None

        # Initialize the table model
        # Note: If enabled, this adds an empty column for row delete buttons
        self.setModel(
            TableModel(
                parent=self,
                data=[col_data + ([""] if self._allow_delete else []) for col_data in data],
                row_keys=row_keys,
                col_keys=self._col_keys + ([""] if self._allow_delete else []),
                col_is_widget=col_is_widget
            )
        )

        # Connect callbacks
        if self._on_cell_edited is not None:
            self.model().dataChanged.connect(self.on_data_changed)  # type: ignore
        if self._on_selection_changed is not None:
            self.selectionModel().selectionChanged.connect(self.on_selection_changed)  # type: ignore

        # Set column style
        for col in range(self.model().columnCount()):
            self.horizontalHeader().setSectionResizeMode(col, QHeaderView.Stretch)

        # Set row style
        for row in range(self.model().rowCount()):
            self.verticalHeader().setSectionResizeMode(row, QHeaderView.ResizeToContents)
        self.verticalHeader().setStyleSheet("""
            QHeaderView::section {
                padding: 0 3px;
            }
        """)

        # Add row delete buttons if desired
        if self._allow_delete:
            for row in range(self.model().rowCount()):
                delete_button = QPushButton2("", "fa.minus", css="border: none; background: palette(window);")
                if self.model().rowCount() > self._row_count_minimum:
                    delete_button.clicked.connect(  # type: ignore
                        partial(self.on_row_deleted, row)
                    )
                else:
                    delete_button.setEnabled(False)
                self.setIndexWidget(self.model().index(row, self.model().columnCount() - 1), delete_button)

        # Add combo boxes if desired
        if self._col_options is not None:
            for row, row_data in enumerate(data):
                for col, col_data in enumerate(row_data):
                    if self._col_options[col] is not None:
                        combobox = QComboBox()

                        for i, text in enumerate(self._col_options[col]):
                            combobox.addItem(text)

                            # Cell is bound to project configuration:
                            if self._col_types is not None:
                                key = list(self._col_types.keys())[col]
                                if text == self.gui.project.get_str(self._row_prefix + key + "_" + str(row)):
                                    combobox.setCurrentIndex(i)

                        combobox.currentIndexChanged.connect(  # type: ignore
                            partial(self.on_combobox_cell_edited, combobox, row, col)
                        )

                        self.setIndexWidget(self.model().index(row, col), combobox)

    # ------------------------------------------------------------------------------------------------------------------

    def on_data_changed(self, top_left_index: QtCore.QModelIndex, _bottom_right_index: QtCore.QModelIndex) -> None:
        """
        Gets called when the data in a selection changed.
        Note: Always assuming only a single cell is selected.

        @param top_left_index: Model index of the top left selection bounds
        @param _bottom_right_index: Model index of the bottom right selection bounds
        """
        row, col = top_left_index.row(), top_left_index.column()
        Debug(self, f".on_cell_changed({row}, {col})")

        data = str(self.model().data(top_left_index))
        try:
            value = float(data.replace(",", "."))
        except ValueError:
            value = 0.0
        value = Format.float_to_str(value)

        if self._on_cell_edited is not None:
            self._on_cell_edited(value, row, col)

    def on_combobox_cell_edited(self, combobox: QComboBox, row: int, column: int) -> None:
        """
        Gets called when a combobox cell has been edited.

        @param combobox: QCombobox
        @param row: Row index
        @param column: Column index
        """
        Debug(self, ".on_combobox_cell_edited()")

        if self._on_cell_edited is not None:
            self._on_cell_edited(combobox.currentText(), row, column)

    def on_row_deleted(self, row: int) -> None:
        """
        Gets called when a row was deleted.

        @param row: Row
        """
        Debug(self, f".on_row_deleted({row})")

        if self._on_row_deleted is not None:
            self._on_row_deleted(row)

        self.select_last_row()

    # ------------------------------------------------------------------------------------------------------------------

    def select_cell(self, row: Optional[int] = None, col: Optional[int] = None) -> None:
        """
        Selects a cell.
        Any parameter may be left set to None in order to load its value from the selection model.

        @param row: Row
        @param col: Column
        """
        index = self.selectionModel().currentIndex()
        if row is None:
            row = row or index.row()
        if col is None:
            col = col or index.column()
        if row == -1 or col == -1:
            Debug(self, ".select_cell(): Skipped for missing selection", warning=True)
            return
        Debug(self, f".select_cell({row}, {col})")
        self.selectionModel().setCurrentIndex(self.model().index(row, col), QItemSelectionModel.SelectCurrent)
        self.scrollTo(index)

    def select_last_row(self, focus: bool = True) -> None:
        """
        Selects the last row.

        @param focus: Enable to focus the table, disable to explicitly clear the visual selection
        """
        if self.model().rowCount() == 0:
            Debug(self, ".select_last_row(): Skipped for empty table", warning=True)
            return

        Debug(self, ".select_last_row()")

        self.blockSignals(True)

        self.select_cell(self.model().rowCount() - 1, 0)

        if focus:
            if not self.hasFocus():
                self.setFocus()
        else:
            self.clearSelection()

        self.blockSignals(False)

    def get_selected_row(self) -> Optional[int]:
        """
        Returns the currently selected row index.

        @return: Index of currently selected row (None if none selected)
        """
        if self.selectionModel().hasSelection():
            return self.selectionModel().currentIndex().row()
        else:
            return None

    def on_selection_changed(self, _selected: QItemSelection, _deselected: QItemSelection) -> None:
        """
        Gets called when the selection changed.

        @param _selected: Currently selected QItemSelection
        @param _deselected: Currently deselected QItemSelection
        """
        if self.signalsBlocked():
            return

        Debug(self, ".on_selection_changed()")

        if self._on_selection_changed is not None:
            self._on_selection_changed()

    def focusInEvent(self, _event: QFocusEvent) -> None:
        """
        Gets called when the table gained focus.
        This highlights the border and selects the last-selected cell (if any).

        @param _event: QFocusEvent
        """
        Debug(self, ".focusInEvent()")

        if self.model().rowCount() > 0:
            self.select_cell()

        self.set_style(border_color=Theme.MainColor, border_width=2)

    def focusOutEvent(self, _event: QFocusEvent) -> None:
        """
        Gets called when the table lost focus, or when a cell is being edited.
        When not editing, this clears the selection, triggering L{on_selection_changed()}.

        @param _event: QFocusEvent
        """
        if self.state() == QAbstractItemView.EditingState:
            Debug(self, ".focusOutEvent(): Ignored in editing mode")
        else:
            self.clearSelection()
            self.set_style(border_color="black", border_width=1)



""" QMessageBox2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from PyQt5.Qt import QSize
from PyQt5.QtWidgets import QMessageBox
from magneticalc.QtWidgets2.QDialog2 import QDialog2
from magneticalc.QtWidgets2.QIconLabel import QIconLabel
from magneticalc.Theme import Theme


class QMessageBox2(QDialog2):
    """ QMessageBox2 class. """

    def __init__(
            self,
            title: str,
            text: str,
            icon,
            buttons,
            default_button
    ) -> None:
        """
        Initializes a message box.

        @param title: Title
        @param text: Text
        @param icon: Icon
        @param buttons: Buttons
        @param default_button: Default button
        """
        QDialog2.__init__(self, title=title)

        self.choice = 0

        # Map QMessageBox icon flag to QtAwesome icon ID and color
        icons_map = {
            QMessageBox.Information:    ("fa.info-circle", Theme.MainColor),
            QMessageBox.Question:       ("fa.question-circle", Theme.QuestionColor),
            QMessageBox.Warning:        ("fa.exclamation-circle", Theme.FailureColor),
        }
        icon_id, color = icons_map.get(icon, "")

        self.addSpacing(10)

        self.addLayout(
            QIconLabel(
                text=text,
                icon=icon_id,
                text_color=Theme.DialogTextColor,
                icon_color=color,
                icon_size=QSize(48, 48)
            )
        )

        self.addSpacing(15)

        # Map QMessageBox button flags to {Text: (Icon, Callback)} pairs as expected by L{QButtons}
        buttons_map = {
            QMessageBox.No      : {"No"     : ("fa.close", lambda: self.accepting(QMessageBox.No))},
            QMessageBox.Discard : {"Discard": ("fa.trash", lambda: self.accepting(QMessageBox.Discard))},
            QMessageBox.Cancel  : {"Cancel" : ("fa.close", lambda: self.accepting(QMessageBox.Cancel))},
            QMessageBox.Ok      : {"Ok"     : ("fa.check", lambda: self.accepting(QMessageBox.Ok))},
            QMessageBox.Save    : {"Save"   : ("fa.save", lambda: self.accepting(QMessageBox.Save))},
            QMessageBox.Yes     : {"Yes"    : ("fa.check", lambda: self.accepting(QMessageBox.Yes))},
        }

        index = 0
        default_button_index = None

        buttons_dict = {}
        for button_flag, button_dict in buttons_map.items():
            if button_flag & buttons:
                buttons_dict.update(button_dict)
                if button_flag == default_button:
                    default_button_index = index
                index += 1

        button_list = self.addButtons(buttons_dict)

        # Set default button
        for i, button in button_list.items():
            if i == default_button_index:
                button.setFocus()

        self.rejected.connect(  # type: ignore
            self.rejecting
        )

        self.show()

    def accepting(self, button_flag: int) -> None:
        """
        Sets the choice and accepts the dialog.
        """
        self.choice = button_flag
        self.accept()

    def rejecting(self) -> None:
        """
        Gets called when the user hits the ESC key.
        """



""" QTextBrowser2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Optional
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QTextBrowser, QSizePolicy


class QTextBrowser2(QTextBrowser):
    """ QTextBrowser2 class. """

    def __init__(
            self,
            html: Optional[str] = None
    ) -> None:
        """
        Initializes QTextBrowser2.

        @param html: HTML
        """
        QTextBrowser.__init__(self)

        self.setOpenExternalLinks(True)
        self.setFocusPolicy(Qt.NoFocus)
        self.setStyleSheet("""
            background: palette(window);
            border: none;
        """)

        if html:
            self.insertHtml(html)

        # Scroll to top
        cursor = self.textCursor()
        cursor.setPosition(0)
        self.setTextCursor(cursor)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)

    def fit_to_contents(self) -> None:
        """
        Fit to contents.
        """
        self.setMinimumHeight(int(self.document().size().height()))



""" ConditionalDecorator module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Callable, Any


class ConditionalDecorator(object):
    """ ConditionalDecorator class. """

    def __init__(self, condition: bool, decorating_func: Callable, **decorating_args: Any) -> None:
        """
        Initializes a conditional decorator.

        @param condition: Condition
        @param decorating_func: Decorating function
        @param decorating_args: Arguments to decorating function
        """
        self.condition = condition
        self.decorating_func = decorating_func
        self.decorating_args = decorating_args

    def __call__(self, func: Callable) -> Callable:
        """
        Gets called when a function is to be decorated.

        @param func: Function to be decorated
        @return: Decorated function
        """
        return self.decorating_func(func, **self.decorating_args) if self.condition else func



""" VisPyCanvas module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
import time
import numpy as np
from typing import Optional
from si_prefix import si_format
from vispy import io, scene, visuals
from vispy.scene.cameras import TurntableCamera
from vispy.visuals.visual import Visual
from magneticalc.Debug import Debug
from magneticalc.Field import Field
from magneticalc.Metric import Metric
from magneticalc.Theme import Theme


class VisPyCanvas(scene.SceneCanvas):
    """ VisPyCanvas class. """

    # Font
    DefaultFontFace = Theme.DefaultFontFace
    DefaultFontSize = 9

    # Enable to additionally debug drawing of visuals
    DebugVisuals = False

    # Enable to additionally debug perspective changes
    DebugPerspective = False

    # Base colors
    White = np.array([1, 1, 1, 1])
    Black = np.array([0, 0, 0, 1])

    # Display settings
    FieldArrowHeadSize = 10
    FieldPointSize = 10
    WirePointSize = 4
    WirePointSelectedSize = 10
    WirePointSelectedColor = (1, 0, 0)

    # Zoom limits
    ScaleFactorMin = 1e-3
    ScaleFactorMax = 1e+4

    # Magnitude limit (mitigating divisions by zero)
    MagnitudeLimit = 1e-12

    # Formatting settings
    MagnitudePrecision = 1

    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Initialize VisPy canvas.

        @param gui: GUI
        """
        scene.SceneCanvas.__init__(self)
        Debug(self, ": Init", init=True)

        self.unfreeze()

        self.gui = gui

        self.redraw_start_time: Optional[float] = None

        self.view_main = self.central_widget.add_view()
        self.view_text = self.view_main.add_view()

        self.view_main.camera = TurntableCamera(
            fov=0,  # Use orthographic projection
            translate_speed=3
        )

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.font_manager = visuals.text.text.FontManager(method="gpu")

        self.visual_perspective_info = scene.visuals.create_visual_node(visuals.TextVisual)(
            parent=self.view_text.scene,
            pos=(10, 10),
            anchor_x="left",
            anchor_y="bottom",
            face=self.DefaultFontFace,
            font_size=self.DefaultFontSize,
            font_manager=self.font_manager
        )

        self.visual_field_labels = []  # See: "create_field_labels()", "delete_field_labels()", "redraw_field_labels()"

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self.visual_coordinate_system = scene.visuals.create_visual_node(visuals.XYZAxisVisual)()

        self.visual_wire_segments = scene.visuals.create_visual_node(visuals.LineVisual)()
        self.visual_wire_points_selected = scene.visuals.create_visual_node(visuals.MarkersVisual)()
        self.visual_wire_points_sliced = scene.visuals.create_visual_node(visuals.MarkersVisual)()
        self.visual_field_points = scene.visuals.create_visual_node(visuals.MarkersVisual)()
        self.visual_field_arrow_lines = scene.visuals.create_visual_node(visuals.LineVisual)()
        self.visual_field_arrow_heads = scene.visuals.create_visual_node(visuals.MarkersVisual)()

        if self.DebugVisuals:
            Debug(self, f": visual_wire_segments        =    {self.visual_wire_segments}")
            Debug(self, f": visual_wire_points_selected = {self.visual_wire_points_selected}")
            Debug(self, f": visual_wire_points_sliced   = {self.visual_wire_points_sliced}")
            Debug(self, f": visual_field_points         = {self.visual_field_points}")
            Debug(self, f": visual_field_arrow_lines    =    {self.visual_field_arrow_lines}")
            Debug(self, f": visual_field_arrow_heads    = {self.visual_field_arrow_heads}")

        self.foreground = None
        self.background = None
        self.update_color_scheme()

        self.initializing = False
        self.visual_startup_info = scene.visuals.create_visual_node(visuals.TextVisual)(
            parent=self.view_text.scene,
            pos=(10, 10 + 2 * self.DefaultFontSize),
            anchor_x="left",
            anchor_y="bottom",
            bold=True,
            text="Performing initial just-in-time compilation;\n"
                 "subsequent calculations will execute faster!\n",
            color=(1, .55, 0),  # Orange
            face=self.DefaultFontFace,
            font_size=self.DefaultFontSize,
            font_manager=self.font_manager
        )

        # Insert perspective change handler
        self.signals_blocked = False
        self.super_perspective_changed = self.view_main.camera.view_changed
        self.view_main.camera.view_changed = self.on_perspective_changed

        self.freeze()

    def update_color_scheme(self) -> None:
        """
        Updates the color scheme.
        """
        Debug(self, ".update_color_scheme()")

        dark_background = self.gui.project.get_bool("dark_background")
        self.foreground = self.White if dark_background else self.Black
        self.background = self.Black if dark_background else self.White
        self.bgcolor = self.background
        self.visual_perspective_info.color = np.append(self.foreground[:3], 0.6)

    def set_visible(self, visual: Visual, is_visible: bool) -> None:
        """
        Sets some visual's visibility.

        @param visual: Visual
        @param is_visible: Visibility
        """
        visual.parent = self.view_main.scene if is_visible else None

    def load_perspective(self, redraw: bool = True) -> None:
        """
        Loads perspective from project.

        @param redraw: Enable to trigger final redraw
        """
        Debug(self, ".load_perspective()")

        self.signals_blocked = True
        self.view_main.camera.azimuth = self.gui.project.get_float("azimuth")
        self.view_main.camera.elevation = self.gui.project.get_float("elevation")
        self.view_main.camera.scale_factor = self.gui.project.get_float("scale_factor")
        self.signals_blocked = False

        if redraw:
            self.on_perspective_changed()
            self.redraw()

    def on_perspective_changed(self) -> None:
        """
        Handles a change of perspective.
        """
        if self.signals_blocked:
            if self.DebugPerspective:
                Debug(self, ".on_perspective_changed(): Skipped")
            return
        else:
            if self.DebugPerspective:
                Debug(self, ".on_perspective_changed()")

        # Limit scale factor
        if self.view_main.camera.scale_factor > self.ScaleFactorMax:
            self.view_main.camera.scale_factor = self.ScaleFactorMax
        elif self.view_main.camera.scale_factor < self.ScaleFactorMin:
            self.view_main.camera.scale_factor = self.ScaleFactorMin

        if self.view_main.camera.azimuth != self.gui.project.get_float("azimuth"):
            self.gui.project.set_float("azimuth", self.view_main.camera.azimuth)
            if self.DebugPerspective:
                Debug(self, f".on_perspective_changed(): azimuth = {self.view_main.camera.azimuth}")

        if self.view_main.camera.elevation != self.gui.project.get_float("elevation"):
            self.gui.project.set_float("elevation", self.view_main.camera.elevation)
            if self.DebugPerspective:
                Debug(self, f".on_perspective_changed(): elevation = {self.view_main.camera.elevation}")

        if self.view_main.camera.scale_factor != self.gui.project.get_float("scale_factor"):
            self.gui.project.set_float("scale_factor", self.view_main.camera.scale_factor)
            if self.DebugPerspective:
                Debug(self, f".on_perspective_changed(): scale_factor = {self.view_main.camera.scale_factor}")

        self.super_perspective_changed()
        self.redraw_perspective_info()

    def redraw_perspective_info(self) -> None:
        """
        Redraws the perspective info text.
        """
        visible = self.gui.project.get_bool("show_perspective_info")

        self.visual_perspective_info.parent = self.view_text.scene if visible else None

        if visible:
            # Calculate linearized zoom value from VisPy "scale factor"
            zoom_log_range = np.log10(self.ScaleFactorMax) - np.log10(self.ScaleFactorMin)
            zoom_log_shift = np.log10(self.view_main.camera.scale_factor) - np.log10(self.ScaleFactorMin)
            zoom = 1000 * (1 - zoom_log_shift / zoom_log_range)

            self.visual_perspective_info.text = \
                f"Azimuth: {self.view_main.camera.azimuth:+4.0f} °   " + \
                f"Elevation: {self.view_main.camera.elevation:+3.0f} °   " + \
                f"Zoom: {zoom:4.0f}"

    def redraw(self) -> None:
        """
        Redraws the canvas.
        """
        Debug(self, ".redraw()")

        self.redraw_start_time = time.monotonic()

        self.update_color_scheme()

        if self.gui.model.valid:
            self.initializing = False
        self.visual_startup_info.parent = self.view_text if self.initializing else None

        self.set_visible(self.visual_coordinate_system, self.gui.project.get_bool("show_coordinate_system"))
        self.redraw_perspective_info()

        self.redraw_wire_segments()
        self.redraw_wire_points_sliced()
        self.redraw_wire_points_selected()

        # Determine which field colors to use (if at all)
        if self.gui.model.metric.valid:
            # Use metric colors
            boost = self.gui.project.get_float("field_boost")
            direction = 1 if self.gui.project.get_bool("dark_background") else -1
            colors = Metric.boost_colors(boost, direction, self.gui.model.metric.colors.copy())
        else:
            if self.gui.model.sampling_volume.valid:
                # Use foreground color for all arrows and points
                colors = [self.foreground] * self.gui.model.sampling_volume.points_count
            else:
                # Dummy argument (not accessed by redraw_field_arrows/_points/_labels in this case)
                colors = None

        self.redraw_field_arrows(colors)
        self.redraw_field_points(colors)
        self.redraw_field_labels(colors)

        redraw_time = time.monotonic() - self.redraw_start_time
        Debug(self, f".redraw(): Finished (took {redraw_time:.2f} s)", success=True)

    # ------------------------------------------------------------------------------------------------------------------

    def redraw_wire_segments(self) -> None:
        """
        Redraws wire segments.
        """
        visible = self.gui.model.wire.valid and self.gui.project.get_bool("show_wire_segments")

        self.set_visible(self.visual_wire_segments, visible)

        if visible:
            if self.DebugVisuals:
                Debug(self, f".redraw_wire_segments(): pos[{len(self.gui.model.wire.points_sliced)}]")

            self.visual_wire_segments.set_data(
                pos=self.gui.model.wire.points_sliced,
                connect="strip",
                color=self.foreground
            )

    def redraw_wire_points_sliced(self) -> None:
        """
        Redraws sliced wire points.
        """
        visible = \
            self.gui.model.wire.valid and \
            self.gui.project.get_bool("show_wire_points")

        self.set_visible(self.visual_wire_points_sliced, visible)

        if visible:
            if self.DebugVisuals:
                Debug(self, f".redraw_wire_points_sliced(): pos[{len(self.gui.model.wire.points_sliced)}]")

            self.visual_wire_points_sliced.set_data(
                pos=self.gui.model.wire.points_sliced,
                face_color=self.foreground,
                size=VisPyCanvas.WirePointSize,
                edge_width=0,
                edge_color=None,
                symbol="disc"
            )

    def redraw_wire_points_selected(self) -> None:
        """
        Redraws selected wire base points.
        """
        point_index = self.gui.sidebar_left.wire_widget.table.get_selected_row()

        visible = \
            self.gui.model.wire.valid and \
            self.gui.project.get_bool("show_wire_points") and \
            point_index is not None

        if visible:
            points_selected = np.array([
                self.gui.model.wire.points_transformed[i]
                for i in range(
                    point_index,
                    len(self.gui.model.wire.points_transformed) + point_index - 1,
                    len(self.gui.model.wire.points_base)
                )
            ])

            if self.DebugVisuals:
                Debug(self, f".redraw_wire_points_selected(): pos[{len(points_selected)}]")

            self.visual_wire_points_selected.set_data(
                pos=points_selected,
                face_color=VisPyCanvas.WirePointSelectedColor,
                size=VisPyCanvas.WirePointSelectedSize,
                edge_width=0,
                edge_color=None,
                symbol="disc"
            )

        self.set_visible(self.visual_wire_points_selected, visible)

    # ------------------------------------------------------------------------------------------------------------------

    def redraw_field_arrows(self, colors: np.ndarray) -> None:
        """
        Redraws field arrows.

        @param colors: Colors
        """
        sampling_volume_resolution = self.gui.model.sampling_volume.resolution
        arrow_head_scale = VisPyCanvas.FieldArrowHeadSize * self.gui.project.get_float("field_arrow_head_scale")
        arrow_line_scale = 2 * (1 / sampling_volume_resolution) * self.gui.project.get_float("field_arrow_line_scale")

        visible = \
            self.gui.model.field.valid and \
            (arrow_head_scale > 0 or arrow_line_scale > 0) and \
            self.gui.sidebar_left.wire_widget.table.get_selected_row() is None

        if visible:

            line_pairs = np.zeros([2 * self.gui.model.sampling_volume.points_count, 3])
            head_points = np.zeros([self.gui.model.sampling_volume.points_count, 3])

            line_pairs, head_points = Field.get_arrows(
                self.gui.model.sampling_volume.points,
                self.gui.model.field.vectors,
                line_pairs,
                head_points,
                arrow_line_scale,
                VisPyCanvas.MagnitudeLimit
            )

            if self.DebugVisuals:
                Debug(self, f".redraw_field_arrows(): arrow lines: pos[{len(line_pairs)}] color[{len(colors)}]")

            self.visual_field_arrow_lines.set_data(
                pos=line_pairs,
                connect="segments",
                color=np.repeat(colors, 2, axis=0)
            )

            if self.DebugVisuals:
                Debug(self, f".redraw_field_arrows(): arrow heads: pos[{len(head_points)}] face_color[{len(colors)}]")

            self.visual_field_arrow_heads.set_data(
                pos=head_points,
                face_color=colors,
                size=arrow_head_scale,
                edge_width=0,
                edge_color=None,
                symbol="diamond"
            )

        self.set_visible(self.visual_field_arrow_lines, visible)
        self.set_visible(self.visual_field_arrow_heads, visible)

    def redraw_field_points(self, colors: np.ndarray) -> None:
        """
        Redraws field points.

        @param colors: Colors
        """
        point_scale = VisPyCanvas.FieldPointSize * self.gui.project.get_float("field_point_scale")

        visible = \
            self.gui.model.sampling_volume.valid and \
            point_scale > 0 and \
            self.gui.sidebar_left.wire_widget.table.get_selected_row() is None

        if visible:
            if self.DebugVisuals:
                Debug(
                    self,
                    ".redraw_field_points(): "
                    f"pos[{self.gui.model.sampling_volume.points_count}] face_color[{len(colors)}]"
                )

            self.visual_field_points.set_data(
                pos=self.gui.model.sampling_volume.points,
                face_color=colors,
                size=point_scale,
                edge_width=0,
                edge_color=None,
                symbol="disc"
            )

        self.set_visible(self.visual_field_points, visible)

    # ------------------------------------------------------------------------------------------------------------------

    def create_field_labels(self) -> None:
        """
        Creates field labels.
        """
        n = len(self.gui.model.sampling_volume.labeled_indices)

        if self.DebugVisuals:
            Debug(self, f".create_field_labels(): Creating {n} labels …")

        show_gauss = self.gui.project.get_bool("show_gauss")
        field_units, field_factor = self.gui.model.field.get_units(show_gauss=show_gauss)

        # Iterate through the labeled sampling volume points
        for i in range(n):

            sampling_volume_point, field_vector_index = self.gui.model.sampling_volume.labeled_indices[i]
            magnitude = np.linalg.norm(self.gui.model.field.vectors[field_vector_index])
            if np.isnan(magnitude):
                text = "NaN"
            else:
                text = si_format(
                    magnitude * field_factor,
                    precision=VisPyCanvas.MagnitudePrecision,
                    exp_format_str="{value}e{expof10} "
                ) + field_units

            visual = scene.visuals.create_visual_node(visuals.TextVisual)(
                parent=None,
                pos=sampling_volume_point,
                face=self.DefaultFontFace,
                font_size=self.DefaultFontSize,
                color=self.foreground,
                text=text,
                font_manager=self.font_manager
            )

            self.visual_field_labels.append(visual)

        if self.DebugVisuals:
            Debug(self, f".create_field_labels(): Created {n} labels")

    def delete_field_labels(self) -> None:
        """
        Deletes the field labels.
        """
        for visual in self.visual_field_labels:
            visual.parent = None

        if self.DebugVisuals:
            Debug(self, f".delete_field_labels(): Deleted {len(self.visual_field_labels)}")

        self.visual_field_labels = []

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def redraw_field_labels(self, colors) -> None:
        """
        Redraws field labels.

        @param colors: Colors
        """
        visible = \
            self.gui.model.metric.valid and \
            self.gui.project.get_bool("display_field_magnitude_labels") and \
            self.gui.sidebar_left.wire_widget.table.get_selected_row() is None

        if visible:

            # Create field labels if necessary
            if len(self.visual_field_labels) == 0:
                self.create_field_labels()

            if self.DebugVisuals:
                Debug(self, f".redraw_field_labels(): Coloring {len(self.visual_field_labels)}")

            # Update label colors
            show_colored_labels = self.gui.project.get_bool("show_colored_labels")
            for i in range(len(self.visual_field_labels)):
                if show_colored_labels:
                    # Use metric color at labeled sampling volume point
                    _, field_vector_index = self.gui.model.sampling_volume.labeled_indices[i]
                    self.visual_field_labels[i].color = np.append(colors[field_vector_index][:3], 1.0)
                else:
                    # Use foreground color for all labels
                    self.visual_field_labels[i].color = self.foreground

        for visual in self.visual_field_labels:
            visual.parent = self.view_main.scene if visible else None

    # ------------------------------------------------------------------------------------------------------------------

    def save_image(self, filename: str) -> None:
        """
        Saves the currently displayed scene to PNG file.

        @param filename: Filename
        """
        Debug(self, f".save_image({filename})")

        io.write_png(filename, self.render())



""" Metric_Presets module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from magneticalc.Norm_Types import *
from typing import Dict
from magneticalc.Assert_Dialog import Assert_Dialog


class Metric_Presets:
    """ Metric_Presets class. """

    # Metric value: Magnitude in XYZ-space (linear)
    Magnitude = {
        "id"        : "Magnitude",
        "norm_type" : NORM_TYPE_RADIUS,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in X-direction (linear)
    MagnitudeX = {
        "id"        : "Magnitude X",
        "norm_type" : NORM_TYPE_RADIUS_X,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in Y-direction (linear)
    MagnitudeY = {
        "id"        : "Magnitude Y",
        "norm_type" : NORM_TYPE_RADIUS_Y,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in Z-direction (linear)
    MagnitudeZ = {
        "id"        : "Magnitude Z",
        "norm_type" : NORM_TYPE_RADIUS_Z,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in XY-plane (linear)
    MagnitudeXY = {
        "id"        : "Magnitude XY",
        "norm_type" : NORM_TYPE_RADIUS_XY,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in XZ-plane (linear)
    MagnitudeXZ = {
        "id"        : "Magnitude XZ",
        "norm_type" : NORM_TYPE_RADIUS_XZ,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in YZ-plane (linear)
    MagnitudeYZ = {
        "id"        : "Magnitude YZ",
        "norm_type" : NORM_TYPE_RADIUS_YZ,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Unipolar Divergence (linear)
    Divergence = {
        "id"        : "Divergence",
        "norm_type" : NORM_TYPE_DIVERGENCE,
        "polarity"  : 0,
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Negative Divergence (linear)
    NegDivergence = {
        "id"        : "Divergence –",
        "norm_type" : NORM_TYPE_DIVERGENCE,
        "polarity"  : -1,
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Positive Divergence (linear)
    PosDivergence = {
        "id"        : "Divergence +",
        "norm_type" : NORM_TYPE_DIVERGENCE,
        "polarity"  : +1,
        "is_log"    : False,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in XYZ-space (logarithmic)
    LogMagnitude = {
        "id"        : "Log Magnitude",
        "norm_type" : NORM_TYPE_RADIUS,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in X-direction (logarithmic)
    LogMagnitudeX = {
        "id"        : "Log Magnitude X",
        "norm_type" : NORM_TYPE_RADIUS_X,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in Y-direction (logarithmic)
    LogMagnitudeY = {
        "id"        : "Log Magnitude Y",
        "norm_type" : NORM_TYPE_RADIUS_Y,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in Z-direction (logarithmic)
    LogMagnitudeZ = {
        "id"        : "Log Magnitude Z",
        "norm_type" : NORM_TYPE_RADIUS_Z,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in XY-plane (logarithmic)
    LogMagnitudeXY = {
        "id"        : "Log Magnitude XY",
        "norm_type" : NORM_TYPE_RADIUS_XY,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric value: Magnitude in XZ-plane (logarithmic)
    LogMagnitudeXZ = {
        "id"        : "Log Magnitude XZ",
        "norm_type" : NORM_TYPE_RADIUS_XZ,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Magnitude in YZ-plane (logarithmic)
    LogMagnitudeYZ = {
        "id"        : "Log Magnitude YZ",
        "norm_type" : NORM_TYPE_RADIUS_YZ,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Unipolar Divergence (logarithmic)
    LogDivergence = {
        "id"        : "Log Divergence",
        "norm_type" : NORM_TYPE_DIVERGENCE,
        "polarity"  : 0,
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Positive Divergence (logarithmic)
    PosLogDivergence = {
        "id"        : "Log Divergence +",
        "norm_type" : NORM_TYPE_DIVERGENCE,
        "polarity"  : +1,
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Negative Divergence (logarithmic)
    NegLogDivergence = {
        "id"        : "Log Divergence –",
        "norm_type" : NORM_TYPE_DIVERGENCE,
        "polarity"  : -1,
        "is_log"    : True,
        "is_angle"  : False,
        "colormap"  : 0  # divergent
    }

    # Metric preset: Angle in XY-plane
    AngleXY = {
        "id"        : "Angle XY",
        "norm_type" : NORM_TYPE_ANGLE_XY,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : True,
        "colormap"  : 1  # cyclic
    }

    # Metric preset: Angle in XZ-plane
    AngleXZ = {
        "id"        : "Angle XZ",
        "norm_type" : NORM_TYPE_ANGLE_XZ,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : True,
        "colormap"  : 1  # cyclic
    }

    # Metric preset: Angle in YZ-plane
    AngleYZ = {
        "id"        : "Angle YZ",
        "norm_type" : NORM_TYPE_ANGLE_YZ,
        "polarity"  : 0,  # currently only used for divergence
        "is_log"    : False,
        "is_angle"  : True,
        "colormap"  : 1  # cyclic
    }

    # ------------------------------------------------------------------------------------------------------------------

    # List of all above presets
    List = [
        Magnitude,
        MagnitudeX,
        MagnitudeY,
        MagnitudeZ,
        MagnitudeXY,
        MagnitudeXZ,
        MagnitudeYZ,
        Divergence,
        PosDivergence,
        NegDivergence,
        LogMagnitude,
        LogMagnitudeX,
        LogMagnitudeY,
        LogMagnitudeZ,
        LogMagnitudeXY,
        LogMagnitudeXZ,
        LogMagnitudeYZ,
        LogDivergence,
        PosLogDivergence,
        NegLogDivergence,
        AngleXY,
        AngleXZ,
        AngleYZ,
    ]

    # ------------------------------------------------------------------------------------------------------------------

    Fallback = Magnitude

    @staticmethod
    def get_by_id(_id_: str) -> Dict:
        """
        Selects a preset by name.

        @param _id_: Preset ID
        @return: Preset parameters
        """
        for preset in Metric_Presets.List:
            if _id_ == preset["id"]:
                return preset

        Assert_Dialog(False, f"Invalid metric preset ID: Defaulting to \"{Metric_Presets.Fallback['id']}\"")
        return Metric_Presets.Fallback



""" Statusbar module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional
from multiprocessing import cpu_count
from sty import fg
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QCheckBox, QComboBox, QProgressBar, QLabel
from magneticalc.QtWidgets2.QPushButton2 import QPushButton2
from magneticalc.Backend_Types import get_jit_enabled
from magneticalc.Debug import Debug
from magneticalc.Theme import Theme


class Statusbar:
    """ Statusbar class. """

    # Used by L{Debug}
    DebugColor = fg.blue

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Initializes statusbar.

        @param gui: GUI
        """
        Debug(self, ": Init", init=True)
        self.gui = gui

        self._valid = True  # Note: Needs to be true initially, otherwise the first L{invalidate()} would have no effect
        self._canceled = False

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # Start button
        self._start_button = QPushButton2(" ⟨F5⟩ ", "fa.play-circle", self.start, css="padding: 3px; font-size: 13px;")

        # Cancel button
        self._cancel_button = QPushButton2("⟨ESC⟩", "fa.stop-circle", self.cancel, css="padding: 3px; font-size: 13px;")

        # Auto-calculation checkbox
        self._auto_calculation_checkbox = QCheckBox("Auto-Calculation")
        self._auto_calculation_checkbox.toggled.connect(self._auto_calculation_changed)  # type: ignore

        # Number-of-cores combobox
        self._cores_combobox = QComboBox()

        if get_jit_enabled():

            # "Auto" setting
            num_cores_auto = max(1, cpu_count() - 1)

            for i in range(0, cpu_count() + 1):
                if i == 0:
                    self._cores_combobox.addItem(
                        f"Auto ({num_cores_auto} Core" + ("s" if num_cores_auto > 1 else "") + ")"
                    )
                else:
                    self._cores_combobox.addItem(f"{i} Core" + ("s" if i > 1 else ""))

            self._cores_combobox.currentIndexChanged.connect(  # type: ignore
                lambda: self.gui.project.set_int("num_cores", self._cores_combobox.currentIndex())
            )

        else:

            self._cores_combobox.addItem("JIT Disabled")
            self._cores_combobox.setEnabled(False)

        # Progress bar
        self._progressbar = QProgressBar()

        # Status text
        self._label = QLabel()

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # Populate statusbar layout
        self.layout = QHBoxLayout()
        self.layout.addWidget(self._start_button)
        self.layout.addSpacing(4)
        self.layout.addWidget(self._cancel_button)
        self.layout.addSpacing(4)
        self.layout.addWidget(self._auto_calculation_checkbox)
        self.layout.addSpacing(-4)
        self.layout.addWidget(self._cores_combobox)
        self.layout.addSpacing(4)
        self.layout.addWidget(self._progressbar)
        self.layout.addSpacing(4)
        self.layout.addWidget(self._label)
        self.layout.setContentsMargins(8, 2, 10, 2)

        container_widget = QWidget()
        container_widget.setLayout(self.layout)
        gui.statusBar().addPermanentWidget(container_widget, stretch=10)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        gui.statusBar().setSizeGripEnabled(False)

    def reload(self) -> None:
        """
        Reloads the statusbar.
        """
        Debug(self, ".reload()", refresh=True)

        self.gui.blockSignals(True)

        self._auto_calculation_checkbox.setChecked(self.gui.project.get_bool("auto_calculation"))
        if 0 > self.gui.project.get_int("num_cores") > cpu_count():
            self.gui.project.set_int("num_cores", 0)

        if get_jit_enabled():
            num_cores = self.gui.project.get_int("num_cores")
            for i in range(0, cpu_count() + 1):
                if i == num_cores:
                    self._cores_combobox.setCurrentIndex(i)

        self.gui.blockSignals(False)

        self.invalidate()

    # ------------------------------------------------------------------------------------------------------------------

    def _auto_calculation_changed(self) -> None:
        """
        Handles changed auto-calculation setting.
        """
        if self.gui.signalsBlocked():
            return

        self._canceled = False

        self.gui.project.set_bool("auto_calculation", self._auto_calculation_checkbox.isChecked())
        if self._auto_calculation_checkbox.isChecked():
            if not self.gui.model.valid:
                self.gui.recalculate()

    # ------------------------------------------------------------------------------------------------------------------

    def arm(self) -> None:
        """
        "Arms" the statusbar before calculation.
        """
        Debug(self, ".arm()")

        self._valid = False
        self._canceled = False

        self._start_button.setEnabled(False)
        self._cancel_button.setEnabled(True)
        self._auto_calculation_checkbox.setEnabled(False)

        if get_jit_enabled():
            self._cores_combobox.setEnabled(False)

        self._progressbar.setValue(0)
        self.set_progressbar_color(fg_color=Theme.MainColor)

    def disarm(self, success: bool) -> None:
        """
        "Disarms" the statusbar after calculation.

        @param success: Reflects calculation success or failure
        """
        Debug(self, f".disarm(success={success})")

        self._valid = success
        self._canceled = not success

        self._start_button.setEnabled(True)
        self._cancel_button.setEnabled(False)
        self._auto_calculation_checkbox.setEnabled(True)

        if get_jit_enabled():
            self._cores_combobox.setEnabled(True)

        self.set_text("Complete." if success else "Canceled!")
        self.set_progressbar_color(fg_color=Theme.SuccessColor if success else Theme.FailureColor)

        if success:
            self._progressbar.setValue(100)

    def invalidate(self) -> None:
        """
        "Invalidates" the statusbar.
        """
        # Ignore when exiting
        if self.gui.exiting:
            return

        # Don't invalidate more than once
        if not self._valid:
            return

        # Don't invalidate after cancellation
        if self._canceled:
            self._canceled = False
            return

        Debug(self, ".invalidate()")

        self._valid = False

        self._start_button.setEnabled(True)
        self._cancel_button.setEnabled(False)
        self._auto_calculation_checkbox.setEnabled(True)

        if get_jit_enabled():
            self._cores_combobox.setEnabled(True)

        self.set_text("Pending Calculation")
        self.set_progressbar_color(fg_color=Theme.LiteColor, bg_color=Theme.LiteColor)
        self._progressbar.setValue(0)

    def start(self) -> None:
        """
        Starts the calculation.
        """
        Debug(self, ".start()")

        self.gui.recalculate()
        self.disarm(True)

    def cancel(self) -> None:
        """
        Cancels the ongoing calculation.
        """
        Debug(self, ".cancel()")

        if self.gui.calculation_thread is not None:
            self.gui.interrupt_calculation()

        self.disarm(False)

    def set_text(self, text: str) -> None:
        """
        Updates the statusbar text.

        @param text: Text
        """
        # Ignore when exiting
        if self.gui.exiting:
            return

        Debug(self, f": {text}")

        self._label.setText(text)

    def set_progress(self, percentage: int) -> None:
        """
        Updates the progress percentage.

        @param percentage: Percentage
        """
        self._progressbar.setValue(percentage)

    def set_progressbar_color(self, fg_color: str, bg_color: Optional[str] = None) -> None:
        """
        Sets the progressbar color & style.

        @param fg_color: Foreground color
        @param bg_color: Background color (optional)
        """
        self._progressbar.setStyleSheet(
            f"""
            QProgressBar
            {{
                border: 1px solid #888888;
                border-radius: 5px;
                color: #ffffff;
                text-align: center;
                padding: 0px;
                height: 23px;
                background-color: {bg_color if bg_color is not None else "#000000"};
                font-size: 13px;
            }}
            QProgressBar::chunk {{
                padding: 0px;
                background-color: {fg_color};
                text-align: center;
            }}"""
        )



""" Perspective_Widget module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Dict
from functools import partial
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout
from magneticalc.QtWidgets2.QGroupBox2 import QGroupBox2
from magneticalc.QtWidgets2.QHLine import QHLine
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.QtWidgets2.QPushButton2 import QPushButton2
from magneticalc.Debug import Debug
from magneticalc.Perspective_Presets import Perspective_Presets
from magneticalc.Theme import Theme


class Perspective_Widget(QGroupBox2):
    """ Perspective_Widget class. """

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Populates the widget.

        @param gui: GUI
        """
        QGroupBox2.__init__(self, "Perspective", color=Theme.DarkColor)
        Debug(self, ": Init", init=True)
        self.gui = gui

        planar_perspective_layout = QVBoxLayout()
        for preset in Perspective_Presets.List:
            button = QPushButton2(preset["id"], "", clicked=partial(self.set_perspective, preset))
            planar_perspective_layout.addWidget(button)
        self.addLayout(planar_perspective_layout)

        self.addWidget(QHLine())
        xyz_hint_layout = QHBoxLayout()
        xyz_hint_layout.addWidget(QLabel2("Axis Colors:", italic=True, color=Theme.LiteColor))
        xyz_hint_layout.addWidget(QLabel2("X", bold=True, color="#cc0000", align_right=True))
        xyz_hint_layout.addWidget(QLabel2("Y", bold=True, color="#00cc00", align_right=True))
        xyz_hint_layout.addWidget(QLabel2("Z", bold=True, color="#0000cc", align_right=True))
        self.addLayout(xyz_hint_layout)

    def reload(self) -> None:
        """
        Reloads the widget.
        """
        Debug(self, ".reload()", refresh=True)

    def set_perspective(self, preset: Dict) -> None:
        """
        Sets display perspective.

        @param preset: Perspective preset (parameters, see VisPyCanvas module)
        """
        Debug(self, ".set_perspective()")
        self.gui.project.set_float("azimuth", preset["azimuth"])
        self.gui.project.set_float("elevation", preset["elevation"])
        self.gui.vispy_canvas.view_main.camera.azimuth = preset["azimuth"]
        self.gui.vispy_canvas.view_main.camera.elevation = preset["elevation"]
        self.gui.redraw()



""" Perspective_Presets module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Dict
import numpy as np
from magneticalc.Assert_Dialog import Assert_Dialog


class Perspective_Presets:
    """ Perspective_Presets class. """

    # Preset: Isometric
    Isometric = {
        "id": "Isometric",
        "azimuth": 135.0,
        "elevation": round(np.arctan(1 / np.sqrt(2)) * 180 / np.pi, 4)
    }

    # Preset: XY-plane
    PlaneXY = {
        "id": "Plane XY",
        "azimuth": 0.0,
        "elevation": 90.0
    }

    # Preset: XZ-plane
    PlaneXZ = {
        "id": "Plane XZ",
        "azimuth": 0.0,
        "elevation": 0.0
    }

    # Preset: YZ-plane
    PlaneYZ = {
        "id": "Plane YZ",
        "azimuth": 90.0,
        "elevation": 0.0
    }

    # ------------------------------------------------------------------------------------------------------------------

    # List of all above presets
    List = [
        Isometric,
        PlaneXY,
        PlaneXZ,
        PlaneYZ
    ]

    # ------------------------------------------------------------------------------------------------------------------

    Fallback = Isometric

    @staticmethod
    def get_by_id(_id_: str) -> Dict:
        """
        Selects a preset by name.

        @param _id_: Preset ID
        @return: Preset parameters
        """
        for preset in Perspective_Presets.List:
            if _id_ == preset["id"]:
                return preset

        Assert_Dialog(False, f"Invalid perspective preset ID: Defaulting to \"{Perspective_Presets.Fallback['id']}\"")
        return Perspective_Presets.Fallback



""" Backend_CUDA module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Callable, Tuple, Optional
import math
import numpy as np
from numba import cuda
from PyQt5.QtCore import QThread
from magneticalc.Constants import Constants
from magneticalc.Debug import Debug
from magneticalc.Field_Types import FIELD_TYPE_A, FIELD_TYPE_B


class Backend_CUDA:
    """
    Implements the Biot-Savart law for calculating the magnetic flux density (B-field) and vector potential (A-field).
    Backend: JIT + CUDA.
    """

    def __init__(
            self,
            field_type: int,
            distance_limit: float,
            length_scale: float,
            dc: float,
            current_elements: np.ndarray,
            sampling_volume_points: np.ndarray,
            sampling_volume_permeabilities: np.ndarray,
            progress_callback: Callable
    ) -> None:
        """
        Initializes the class attributes.

        @param field_type: Field type
        @param distance_limit: Distance limit (mitigating divisions by zero)
        @param length_scale: Length scale (m)
        @param dc: Wire current (A)
        @param current_elements: Ordered list of current elements (pairs: [element center, element direction])
        @param sampling_volume_points: Ordered list of sampling volume points
        @param sampling_volume_permeabilities: Ordered list of sampling volume's relative permeabilities µ_r
        @param progress_callback: Progress callback
        """
        Debug(self, ": Init", init=True)

        self._field_type = field_type
        self._distance_limit = distance_limit
        self._length_scale = length_scale
        self._dc = dc
        self._current_elements = current_elements
        self._sampling_volume_points = sampling_volume_points
        self._sampling_volume_permeabilities = sampling_volume_permeabilities
        self._progress_callback = progress_callback

    @staticmethod
    @cuda.jit
    def worker(
            field_type: int,
            distance_limit: float,
            length_scale: float,
            element_centers: np.ndarray,
            element_directions: np.ndarray,
            sampling_volume_points: np.ndarray,
            sampling_volume_permeabilities: np.ndarray,
            field_vectors: np.ndarray,
            total_calculations: np.ndarray,
            total_skipped_calculations: np.ndarray
    ) -> None:
        """
        Applies the Biot-Savart law for calculating the magnetic flux density (B-field) or vector potential (A-field)
        for all sampling volume points.

        @param field_type: Field type
        @param distance_limit: Distance limit (mitigating divisions by zero)
        @param length_scale: Length scale (m)
        @param element_centers: Ordered list of current elements centers
        @param element_directions: Ordered list of current elements directions
        @param sampling_volume_points: Sampling volume points
        @param sampling_volume_permeabilities: Ordered list of sampling volume's relative permeabilities µ_r
        @param field_vectors: Field vectors (output array)
        @param total_calculations: Total number of calculations (output array)
        @param total_skipped_calculations: Total number of skipped calculations (output array)
        """

        sampling_volume_index = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x  # type: ignore

        if sampling_volume_index >= sampling_volume_points.shape[0]:
            return

        total_calculations[sampling_volume_index] = 0
        total_skipped_calculations[sampling_volume_index] = 0

        vector_x = 0
        vector_y = 0
        vector_z = 0

        for current_element_index in range(element_centers.shape[0]):
            vector_distance_x = (sampling_volume_points[sampling_volume_index][0] -
                                 element_centers[current_element_index][0]) * length_scale
            vector_distance_y = (sampling_volume_points[sampling_volume_index][1] -
                                 element_centers[current_element_index][1]) * length_scale
            vector_distance_z = (sampling_volume_points[sampling_volume_index][2] -
                                 element_centers[current_element_index][2]) * length_scale

            # Calculate distance (mitigating divisions by zero)
            scalar_distance = math.sqrt(vector_distance_x ** 2 + vector_distance_y ** 2 + vector_distance_z ** 2)
            if scalar_distance < distance_limit:
                scalar_distance = distance_limit
                total_skipped_calculations[sampling_volume_index] += 1

            total_calculations[sampling_volume_index] += 1

            if field_type == FIELD_TYPE_A:

                # Calculate A-field (vector potential)
                vector_x += element_directions[current_element_index][0] * length_scale / scalar_distance
                vector_y += element_directions[current_element_index][1] * length_scale / scalar_distance
                vector_z += element_directions[current_element_index][2] * length_scale / scalar_distance

            elif field_type == FIELD_TYPE_B:

                # Calculate B-field (flux density)
                a_1 = element_directions[current_element_index][0] * length_scale
                a_2 = element_directions[current_element_index][1] * length_scale
                a_3 = element_directions[current_element_index][2] * length_scale
                vector_x += (a_2 * vector_distance_z - a_3 * vector_distance_y) / (scalar_distance ** 3)
                vector_y += (a_3 * vector_distance_x - a_1 * vector_distance_z) / (scalar_distance ** 3)
                vector_z += (a_1 * vector_distance_y - a_2 * vector_distance_x) / (scalar_distance ** 3)

        field_vectors[sampling_volume_index, 0] = vector_x * sampling_volume_permeabilities[sampling_volume_index]
        field_vectors[sampling_volume_index, 1] = vector_y * sampling_volume_permeabilities[sampling_volume_index]
        field_vectors[sampling_volume_index, 2] = vector_z * sampling_volume_permeabilities[sampling_volume_index]

    def get_result(self) -> Optional[Tuple[int, int, np.ndarray]]:
        """
        Calculates the field at every point of the sampling volume.

        @return: (Total # of calculations, total # of skipped calculations, field) if successful, None if interrupted
        """
        Debug(self, ".get_result()")

        element_centers = [element[0] for element in self._current_elements]
        element_directions = [element[1] for element in self._current_elements]

        element_centers_global = cuda.to_device(element_centers)
        element_directions_global = cuda.to_device(element_directions)

        total_calculations = 0
        total_skipped_calculations = 0
        field_vectors = np.zeros(shape=(0, 3))

        # Split the calculation into chunks for progress update and interruption handling
        chunk_size_max = 1024 * 16
        chunk_start = 0
        remaining = len(self._sampling_volume_points)

        while remaining > 0:

            if remaining >= chunk_size_max:
                chunk_size = chunk_size_max
            else:
                chunk_size = remaining

            sampling_volume_points_global = cuda.to_device(
                self._sampling_volume_points[chunk_start:chunk_start + chunk_size]
            )
            sampling_volume_permeabilities_global = cuda.to_device(
                self._sampling_volume_permeabilities[chunk_start:chunk_start + chunk_size]
            )

            # Signal progress update, handle interrupt
            self._progress_callback(100 * chunk_start / len(self._sampling_volume_points))

            if QThread.currentThread().isInterruptionRequested():
                Debug(self, ".get_result(): WARNING: Interruption requested, exiting now", warning=True)
                return None

            remaining -= chunk_size
            chunk_start += chunk_size

            total_calculations_global = cuda.to_device(np.zeros(chunk_size))
            total_skipped_calculations_global = cuda.to_device(np.zeros(chunk_size))
            field_vectors_global = cuda.device_array((chunk_size, 3))

            TPB = 1024   # Maximum threads per block
            BPG = 65536  # Maximum blocks per grid

            Backend_CUDA.worker[BPG, TPB](  # type: ignore
                self._field_type,
                self._distance_limit,
                self._length_scale,
                element_centers_global,
                element_directions_global,
                sampling_volume_points_global,
                sampling_volume_permeabilities_global,
                field_vectors_global,
                total_calculations_global,
                total_skipped_calculations_global
            )

            total_calculations_local = total_calculations_global.copy_to_host()
            total_skipped_calculations_local = total_skipped_calculations_global.copy_to_host()
            field_vectors_local = field_vectors_global.copy_to_host()

            if self._field_type == FIELD_TYPE_A or self._field_type == FIELD_TYPE_B:
                # Field is A-field or B-field
                field_vectors_local = field_vectors_local * self._dc * Constants.mu_0 / 4 / np.pi

            total_calculations += int(sum(total_calculations_local))
            total_skipped_calculations += int(sum(total_skipped_calculations_local))
            field_vectors = np.append(field_vectors, field_vectors_local, axis=0)

        self._progress_callback(100)

        return total_calculations, total_skipped_calculations, np.array(field_vectors)



""" Model access module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from types import TracebackType
from typing import Optional, Type
from sty import fg
from magneticalc.Debug import Debug


class ModelAccess:
    """ Model access class. """

    # Used by L{Debug}
    DebugColor = fg.yellow

    # Debug validity of hierarchy levels
    DebugValidity = False

    def __init__(
            self,
            gui: GUI,  # type: ignore
            recalculate: bool
    ) -> None:
        """
        Initializes model access.

        @param gui: GUI
        @param recalculate: Enable to recalculate upon exiting the context
        """
        self.gui = gui
        self._recalculate = recalculate

    def __enter__(self) -> None:
        """
        Entering the context kills possibly running calculation if recalculation is enabled.
        """
        Debug(self, ".enter()")

        if self._recalculate:
            if self.gui.calculation_thread is not None:
                self.gui.interrupt_calculation()

    def __exit__(
            self,
            _exc_type: Optional[Type[BaseException]],
            _exc_val: Optional[BaseException],
            _exc_tb: Optional[TracebackType]
    ) -> None:
        """
        Leaving the context starts recalculation if enabled; otherwise, just redraw.
        """
        Debug(self, ".exit()")

        if self.DebugValidity:
            for obj in [
                self.gui.model.wire,
                self.gui.model.sampling_volume,
                self.gui.model.field,
                self.gui.model.metric,
                self.gui.model.parameters
            ]:
                Debug(self, f": {type(obj).__name__:>14}.valid = {obj.valid}", success=obj.valid, warning=not obj.valid)

        if not self.gui.model.valid:
            self.gui.invalidate_statusbar.emit()

        if self._recalculate:
            self.gui.re_wrapper()



""" Constraint module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from magneticalc.Norm_Types import *
from typing import Union, List
import numpy as np
from magneticalc.Metric import metric_norm
from magneticalc.Comparison_Types import *


class Constraint:
    """ Constraint class. """

    """ Supported norm types. """
    Norm_Types_Supported = [
        NORM_TYPE_X,
        NORM_TYPE_Y,
        NORM_TYPE_Z,
        NORM_TYPE_RADIUS,
        NORM_TYPE_RADIUS_X,
        NORM_TYPE_RADIUS_Y,
        NORM_TYPE_RADIUS_Z,
        NORM_TYPE_RADIUS_XY,
        NORM_TYPE_RADIUS_XZ,
        NORM_TYPE_RADIUS_YZ,
        NORM_TYPE_ANGLE_XY,
        NORM_TYPE_ANGLE_XZ,
        NORM_TYPE_ANGLE_YZ
    ]

    """ Supported norm types using minimum and maximum angles in degrees. """
    Norm_Types_Supported_Degrees = [
        NORM_TYPE_ANGLE_XY,
        NORM_TYPE_ANGLE_XZ,
        NORM_TYPE_ANGLE_YZ
    ]

    """ Supported comparison types. """
    Comparison_Types_Supported = list(Comparison_Types_Names_Map.keys())

    def __init__(self, norm_type: int, comparison_type: int, _min: float, _max: float, permeability: float):
        """
        Initializes the constraint.

        @param norm_type: Norm type
        @param comparison_type: Comparison type
        @param _min: Minimum value
        @param _max: Maximum value
        @param permeability: Relative permeability µ_r
        """
        norm_type = norm_type_safe(norm_type)
        comparison_type = comparison_type_safe(comparison_type)

        self._norm_type = norm_type
        self._comparison_type = comparison_type

        self._min = _min
        self._max = _max

        if norm_type in self.Norm_Types_Supported_Degrees:
            # Clip angles to range
            self._min %= 360
            self._max %= 360

            # Normalize angles
            self._min /= 360.0
            self._max /= 360.0

            # Ensure monotonicity
            if self._min > self._max:
                self._min, self._max = self._max, self._min

        self.permeability = permeability

    def evaluate(self, point: Union[np.ndarray, List[float]]) -> bool:
        """
        Evaluate this constraint at some point.

        @param point: Point (3D vector)
        """
        norm = metric_norm(self._norm_type, point)

        if self._comparison_type == COMPARISON_TYPE_IN_RANGE:

            return self._min <= norm <= self._max

        elif self._comparison_type == COMPARISON_TYPE_NOT_IN_RANGE:

            return not self._min <= norm <= self._max

        else:
            # Invalid comparison ID
            return False



""" Parameters_Widget module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
import numpy as np
from si_prefix import si_format
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout
from magneticalc.QtWidgets2.QGroupBox2 import QGroupBox2
from magneticalc.QtWidgets2.QLabel2 import QLabel2
from magneticalc.Debug import Debug
from magneticalc.Field_Types import FIELD_TYPE_A, FIELD_TYPE_B
from magneticalc.Theme import Theme


class Parameters_Widget(QGroupBox2):
    """ Parameters_Widget class. """

    # Formatting settings
    ValuePrecision = 1

    def __init__(
            self,
            gui: GUI  # type: ignore
    ) -> None:
        """
        Populates the widget.

        @param gui: GUI
        """
        QGroupBox2.__init__(self, "Parameters", color=Theme.DarkColor)
        Debug(self, ": Init", init=True)
        self.gui = gui

        self.gui.model.set_parameters(invalidate=False)

        results_layout = QHBoxLayout()
        results_left = QVBoxLayout()
        results_middle = QVBoxLayout()
        results_right = QVBoxLayout()
        results_layout.addLayout(results_left)
        results_layout.addLayout(results_middle)
        results_layout.addLayout(results_right)
        self.addLayout(results_layout)

        results_left.addWidget(QLabel2("Wire length:"))
        self.wire_length_value_label = QLabel2("", color=Theme.MainColor, align_right=True)
        results_middle.addWidget(self.wire_length_value_label)
        self.wire_length_units_label = QLabel2("N/A", color=Theme.MainColor, expand=False)
        results_right.addWidget(self.wire_length_units_label)

        results_left.addWidget(QLabel2("Magnetic Dipole Moment:"))
        self.magnetic_dipole_moment_value_label = QLabel2("", color=Theme.MainColor, align_right=True)
        results_middle.addWidget(self.magnetic_dipole_moment_value_label)
        self.magnetic_dipole_moment_units_label = QLabel2("N/A", color=Theme.MainColor, expand=False)
        results_right.addWidget(self.magnetic_dipole_moment_units_label)

        results_left.addWidget(QLabel2("Energy:"))
        self.energy_value_label = QLabel2("", color=Theme.MainColor, align_right=True)
        results_middle.addWidget(self.energy_value_label)
        self.energy_units_label = QLabel2("N/A", color=Theme.MainColor, expand=False)
        results_right.addWidget(self.energy_units_label)

        results_left.addWidget(QLabel2("Self-inductance:"))
        self.self_inductance_value_label = QLabel2("", color=Theme.MainColor, align_right=True)
        results_middle.addWidget(self.self_inductance_value_label)
        self.self_inductance_units_label = QLabel2("N/A", color=Theme.MainColor, expand=False)
        results_right.addWidget(self.self_inductance_units_label)

        self.update()

    def reload(self) -> None:
        """
        Reloads the widget.
        """
        Debug(self, ".reload()", refresh=True)

    # ------------------------------------------------------------------------------------------------------------------

    def update(self) -> None:
        """
        Updates the widget.
        """
        Debug(self, ".update()", refresh=True)

        self.update_labels()
        self.update_controls()

    def update_labels(self) -> None:
        """
        Updates the labels.
        """
        Debug(self, ".update_labels()", refresh=True)

        if self.gui.model.parameters.valid:

            self.wire_length_value_label.set(
                f"{self.gui.model.wire.length:.2f}", color=Theme.MainColor, bold=True
            )
            self.wire_length_units_label.set("cm", color=Theme.MainColor, bold=True)

            if self.gui.model.field.type == FIELD_TYPE_A:

                self.energy_value_label.set("", color=Theme.MainColor)
                self.energy_units_label.set("N/A", color=Theme.MainColor)

                self.self_inductance_value_label.set("", color=Theme.MainColor)
                self.self_inductance_units_label.set("N/A", color=Theme.MainColor)

                self.magnetic_dipole_moment_value_label.set("", color=Theme.MainColor)
                self.magnetic_dipole_moment_units_label.set("N/A", color=Theme.MainColor)

            elif self.gui.model.field.type == FIELD_TYPE_B:

                energy_value = self.gui.model.parameters.energy
                if np.isnan(energy_value):
                    energy_value = "NaN NaN"
                else:
                    energy_value = si_format(
                        energy_value,
                        precision=self.ValuePrecision,
                        exp_format_str="{value}e{expof10} "
                    ) + "J"
                self.energy_value_label.set(energy_value.split(" ")[0], color=Theme.MainColor, bold=True)
                self.energy_units_label.set(energy_value.split(" ")[1], color=Theme.MainColor, bold=True)

                self_inductance_value = self.gui.model.parameters.self_inductance
                if np.isnan(self_inductance_value):
                    self_inductance_value = "NaN NaN"
                else:
                    self_inductance_value = si_format(
                        self_inductance_value,
                        precision=self.ValuePrecision,
                        exp_format_str="{value}e{expof10} "
                    ) + "H"
                self.self_inductance_value_label.set(
                    self_inductance_value.split(" ")[0], color=Theme.MainColor, bold=True
                )
                self.self_inductance_units_label.set(
                    self_inductance_value.split(" ")[1], color=Theme.MainColor, bold=True
                )

                magnetic_dipole_moment_value = self.gui.model.parameters.magnetic_dipole_moment
                if np.isnan(magnetic_dipole_moment_value):
                    magnetic_dipole_moment_value = "NaN NaN"
                else:
                    magnetic_dipole_moment_value = si_format(
                        magnetic_dipole_moment_value,
                        precision=self.ValuePrecision,
                        exp_format_str="{value}e{expof10} "
                    ) + "A·m²"
                self.magnetic_dipole_moment_value_label.set(
                    magnetic_dipole_moment_value.split(" ")[0], color=Theme.MainColor, bold=True
                )
                self.magnetic_dipole_moment_units_label.set(
                    magnetic_dipole_moment_value.split(" ")[1], color=Theme.MainColor, bold=True
                )

        else:

            self.wire_length_value_label.set("", color=Theme.MainColor)
            self.wire_length_units_label.set("N/A", color=Theme.MainColor)

            self.energy_value_label.set("", color=Theme.MainColor)
            self.energy_units_label.set("N/A", color=Theme.MainColor)

            self.self_inductance_value_label.set("", color=Theme.MainColor)
            self.self_inductance_units_label.set("N/A", color=Theme.MainColor)

            self.magnetic_dipole_moment_value_label.set("", color=Theme.MainColor)
            self.magnetic_dipole_moment_units_label.set("N/A", color=Theme.MainColor)

    def update_controls(self) -> None:
        """
        Updates the controls.
        """
        Debug(self, ".update_controls()", refresh=True)

        self.set_color(Theme.MainColor if self.gui.model.parameters.valid else Theme.FailureColor)



""" QLabel2 module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Optional, Union, Tuple
import qtawesome as qta
from PyQt5.Qt import QFont
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtWidgets import QLabel, QSizePolicy


class QLabel2(QLabel):
    """ QLabel2 class. """

    DebugLabels = False

    def __init__(
            self,
            text: str,
            icon: Optional[str] = None,
            icon_color: Optional[str] = None,
            icon_size: QSize = QSize(16, 16),
            expand: bool = True,
            align_right: bool = False,
            width: Optional[int] = None,
            **kwargs,
    ) -> None:
        """
        Initializes a QLabel.

        Please see L{QLabel2.set()} for supported arguments.

        @param text: Text
        @param icon: QtAwesome icon ID (optional)
        @param icon_color: Icon color (optional)
        @param icon_size: Icon size
        @param expand: Enable to expand
        @param align_right: Enable to align right
        @param width: Width (optional)
        """
        QLabel.__init__(self)

        if icon is not None:
            self.setPixmap(qta.icon(icon, color=icon_color).pixmap(icon_size))

        if expand:
            self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        else:
            self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        if align_right:
            self.setAlignment(Qt.AlignRight | Qt.AlignVCenter)  # type: ignore
        else:
            self.setAlignment(Qt.AlignVCenter)

        if width:
            self.setFixedWidth(width)

        self.set(text, **kwargs)

    def set(
            self,
            text: str,
            font: Optional[QFont] = None,
            font_size: Optional[str] = None,
            bold: bool = False,
            italic: bool = False,
            color: Union[str, Tuple[int, int, int]] = "black",
            css: Optional[str] = None
    ) -> None:
        """
        Sets the label properties.

        @param text: Text
        @param font: QFont
        @param font_size: Native font size
        @param bold: Enable for bold text
        @param italic: Enable for italic text
        @param color: Text color
        @param css: Additional CSS (optional)
        """
        self.setText(text)

        if font:
            self.setFont(font)

        stylesheet_map = {
            f"font-size: {font_size}": font_size,
            "font-weight: bold": bold,
            "font-weight: italic": italic,
            f"color: {color}": color,
            "background-color: red": self.DebugLabels
        }

        self.setStyleSheet(
            (f"{css};" if css is not None else "") +
            ";".join([string for string, condition in stylesheet_map.items() if condition])
        )



""" QIconLabel module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Optional
from PyQt5.Qt import QFont
from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QHBoxLayout
from magneticalc.QtWidgets2.QLabel2 import QLabel2


class QIconLabel(QHBoxLayout):
    """ QIconLabel class. """

    # Default spacing
    HorizontalSpacing = 1

    def __init__(
            self,
            text: str,
            icon: str,
            color: Optional[str] = None,
            text_color: Optional[str] = None,
            icon_color: Optional[str] = None,
            icon_size: QSize = QSize(16, 16),
            font: Optional[QFont] = None,
            expand: bool = True
    ) -> None:
        """
        Initializes the icon label.

        @param text: Label text
        @param icon: QtAwesome icon ID
        @param color: Text and icon color (optional)
        @param text_color: Text color (optional, overrides color)
        @param icon_color: Icon color (optional, overrides color)
        @param icon_size: Icon size
        @param font: QFont
        @param expand: Enable to expand label
        """
        QHBoxLayout.__init__(self)
        self.setContentsMargins(0, 0, 0, 0)

        if color is not None:

            if text_color is None:
                text_color = color

            if icon_color is None:
                icon_color = color

        if icon != "":
            self.addWidget(QLabel2("", icon=icon, icon_color=icon_color, icon_size=icon_size, expand=False))
            self.addSpacing(self.HorizontalSpacing)

        self.addWidget(QLabel2(text, font=font, bold=True, color=text_color, expand=expand))



""" Backend_Types module. """

#  ISC License
#
#  Copyright (c) 2020–2022, Paul Wilhelm <anfrage@paulwilhelm.de>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
from numba import cuda
from magneticalc.Assert_Dialog import Assert_Dialog


def get_jit_enabled() -> bool:
    """
    Checks if JIT is enabled (or at least not explicitly disabled).

    @return: True if JIT enabled, False otherwise
    """
    return (os.environ["NUMBA_DISABLE_JIT"] == "0") if "NUMBA_DISABLE_JIT" in os.environ else True


def get_cuda_available() -> bool:
    """
    Checks if CUDA is available

    @return: True if CUDA available, False otherwise
    """
    return cuda.is_available()


""" Backend type: JIT. """
BACKEND_TYPE_JIT = 0

""" Backend type: CUDA. """
BACKEND_TYPE_CUDA = 1


""" Map of backend types to their availability. """
Backend_Types_Available = {
    BACKEND_TYPE_JIT :  get_jit_enabled(),
    BACKEND_TYPE_CUDA:  get_jit_enabled() and get_cuda_available()
}


""" Map of backend types to names. """
Backend_Types_Names_Map = {
    BACKEND_TYPE_JIT:   "JIT",
    BACKEND_TYPE_CUDA:  "JIT + CUDA"
}

""" Default backend type. """
Backend_Type_Default = BACKEND_TYPE_JIT


def backend_type_safe(backend_type: int) -> int:
    """
    A valid backend type is passed through, but an invalid backend type converts to the default type.

    @param backend_type: Backend type
    @return: Safe backend type
    """
    if backend_type in Backend_Types_Names_Map:
        return backend_type
    else:
        Assert_Dialog(False, "Invalid backend type, using default type")
        return Backend_Type_Default


def backend_type_to_name(backend_type: int) -> str:
    """
    Converts a backend type to a backend name.

    @param backend_type: Backend type
    @return: Backend name
    """
    return Backend_Types_Names_Map[backend_type_safe(backend_type)]
